//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
/* globals BoxState : true, BellCode, CommuatorState, Subclass, RomseyMessageBox */
/* jshint esnext: true */
function BoxState(description) {
    this._description = description;
}
BoxState.prototype = {
    getTooltip: function() { return this._description; }
};

BoxState.NONE = new BoxState("Idle");
BoxState.RECEIVING = new BoxState("Expecting a bell code");
BoxState.AWAITING_TRAIN_ENTERING = new BoxState("Waiting for a 'Train Entering Section' bell code");
BoxState.AWAITING_TRAIN_ENTERING_POSSIBLE_EXIT = new BoxState("Waiting for either 'Train Entering Section' or 'Train Out Of Section' bell code");
BoxState.AWAITING_ATTN_ACK = new BoxState("Waiting for the acknowledgement of an 'Attention' bell code (so they can offer a train)");
BoxState.AWAITING_OFFER_ACK = new BoxState("Waiting for an acknowledgment of their offer of a train");
BoxState.AWAITING_THEIR_LINE_CLEAR = new BoxState("Waiting for the line to Romsey to be LINE CLEAR");
BoxState.PAUSE_TRAIN_START = new BoxState("Arranging for a train to signalled through to Romsey");
BoxState.PAUSE_AFTER_EXIT_ACK = new BoxState("Pausing after acknowledging a 'Train Out Of Section' bell code");
BoxState.EXPECTING_TRAIN = new BoxState("Expecting a train to arrive from Romsey");
BoxState.AWAITING_ENTERING_ACK = new BoxState("Waiting for an acknowledgement of 'Train Entering Section' bell code");
BoxState.AWAITING_THEIR_TRAIN_ON_LINE = new BoxState("Waiting for the line to Romsey to be TRAIN_ON_LINE");
BoxState.AWAITING_EXIT_ATTN_ACK = new BoxState("Waiting for an acknowledgement of an ' Attention' bell code (so they can send 'Train Out Of Section')");
BoxState.AWAITING_OUT_OF_SECTION_ACK = new BoxState("Waiting for an acknowledgement of their 'Train Out Of Section' bell code");
BoxState.AWAITING_SHUNT_ATTN = new BoxState("Waiting for an attn that will signal end of 'Shunt' bell code");
BoxState.AWAITING_END_SHUNT = new BoxState("Waiting for an 'End Shunt' bell code");

function SignalBox(controller, bell) {
    this._controller = controller;
    this._scheduler = controller.getScheduler();
    this._messagebox = controller.getMessageBox();
    this._name = bell.getName();
    this._bell = bell;
    this._state = BoxState.NONE;
    this._preOfferState = BoxState.NONE;
    this._bell.setTooltip(this._state.getTooltip());
    this._offeredTrains = []; // The list of trains offered to the next box, but not yet out of section
    this._trainsIntoOurSection = 0;
    this._trainsLeftOurSection = 0;
    this._idleStarted = this._scheduler.getClock().getTicks();
}

SignalBox.pauseTrainDelay = 5;
SignalBox.timeoutDelay = 60;

SignalBox.prototype = {
    changeState: function(state) {
        if (state !== this._state) {
            this._state = state;
            this._bell.setTooltip(state.getTooltip());
        }
    },
    changeAfterPause: function() {
        if (this._state === BoxState.PAUSE_AFTER_EXIT_ACK) {
            this.changeState(BoxState.NONE);
        }
    },
    startIdleTimer: function() {
        this._idleStarted = this._scheduler.getClock().getTicks();
    },
    trainIntoSection: function() {
        this._trainsIntoOurSection++;
    },
    trainLeftSection: function() {
        this._trainsLeftOurSection++;
    },
    changePrevState: function(state) {
        this._prevState = state;
    },
    _scheduleTimeoutAck: function() {
        this._scheduler.addGameTimeEvent(SignalBox.timeoutDelay,this.timeoutAck,this,this._state);
    },
    timeoutAck: function(oldState) {
        if (oldState === this._state) {
            if (this._bell.isInBellCode()) {
                this._scheduler.addRealTimeEvent(5,this.timeoutAck,this,this._state); // Give them some (real) time to complete their tapping
            } else {
                var reschedule = false;
                var bellcode = null;
                // The state hasn't changed, so we haven't had the expected ack - we need to repeat our bell code
                switch(oldState) {
                    case BoxState.AWAITING_OFFER_ACK:
                    case BoxState.AWAITING_THEIR_LINE_CLEAR:
                        this.changeState(BoxState.AWAITING_ATTN_ACK);
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_ATTN_ACK:
                    case BoxState.AWAITING_EXIT_ATTN_ACK:
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_ENTERING_ACK:
                        bellcode = BellCode.EnteringSection;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_THEIR_TRAIN_ON_LINE:
                        this.changeState(BoxState.AWAITING_ENTERING_ACK);
                        bellcode = BellCode.EnteringSection;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_OUT_OF_SECTION_ACK:
                        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    default:
                        // Unexpected, but we'll try to recover to by just ignoring this timeout event
                        break;
                }
                if (bellcode) {
                    this._bell.scheduleBellCode(bellcode,0);
                }
                if (reschedule) {
                    this._scheduleTimeoutAck();
                }
            }
        }
    },
    sendOutOfOurSection: function() {
        if (this._state === BoxState.PAUSE_TRAIN_START) {
            this._bell.scheduleBellCode(BellCode.EnteringSection,5);
            this.changeState(BoxState.AWAITING_ENTERING_ACK);
            this._scheduleTimeoutAck();
        }
    },
    getName: function() {
        return this._name;
    },
    getBell: function() {
        return this._bell;
    },
    getState: function() {
        return this._state;
    },
    scheduleBellCode: function(code,offset) {
        this._bell.scheduleBellCode(code,offset);
    },
    offerTrain: function(traincode,bellcode,whatToCallOnceTrainOnLine,whoToCallOnceTrainOnLine,whyToProvideOnceTrainOnLine) {
        if (this.canTrainBeOffered()) { // Note: this method must be provided by the actual signal box objects once their construction is complete
            this._preOfferState = this._state;
            this.changeState(BoxState.AWAITING_ATTN_ACK);
            this._trainToOffer = bellcode;
            this._whatToCallOnceTrainOnLine = whatToCallOnceTrainOnLine;
            this._whoToCallOnceTrainOnLine = whoToCallOnceTrainOnLine;
            this._whyToProvideOnceTrainOnLine = whyToProvideOnceTrainOnLine;
            this._bell.scheduleBellCode(BellCode.Attn,1);
            this._scheduleTimeoutAck();
            this._offeredTrains.push(traincode);
            return true;
        } else {
            return false;
        }
    },
    makeTrainOnLineCall: function() {
        this._scheduler.addRealTimeEvent(0.1,this._whatToCallOnceTrainOnLine,this._whoToCallOnceTrainOnLine,this.whyToProvideOnceTrainOnLine);
    },
    makeTrainExitedCall: function() {
        this._scheduler.addRealTimeEvent(0.1,this._whatToCallOnceTrainExited,this._whoToCallOnceTrainExited,this.whyToProvideOnceTrainExited);
    },
    actOnBellPush: function(pushedBellCode) {
        throw new TypeError("actOnBellCode not overridden");
    },
    exitTrain: function(whatToCallOnceTrainExited,whoToCallOnceTrainExited,whyToProvideOnceTrainExited) {
        throw new TypeError("exitTrain not overridden");
    },
    canTrainBeOffered: function() {
        throw new TypeError("canTrainBeOffered not overridden");
    },
    _trainCheckCounts: function(traincode) {
        if (this._trainsIntoOurSection !== this._trainsLeftOurSection) {
            this._messagebox.showMessage(this.getName(),"This is "+this.getName()+" Signaller. Train "+traincode+" has just arrived at my "+
                    "box, but I didn't receive a 'Train Entering Section' bell code from you. If you can send that bell code we can get "+
                    "back to normal operation.");
        }
        this._scheduler.addGameTimeEvent(300,this._trainCheckCounts,this,traincode);
    },
    trainJustArrived: function(traincode) {
        this.trainLeftSection();
        this._trainCheckCounts(traincode);
    },
    trainDepartedSection(traincode) {
        var complained = false;
        for(var i=0; i<this._offeredTrains.length; i++) {
            if (traincode === this._offeredTrains[i]) {
                this._messagebox.showMessage(this.getName(),"This is "+this.getName()+" Signaller. I've not had a 'Train Out "+
                        "Of Section' bell code from you for train "+traincode+". If the train has arrived can you send that bell code so "+
                        "we can get back to normal operation");
            }
            complained = true;
        }
        if (complained) {
            this._scheduler.addGameTimeEvent(300,this.trainDepartedSection,this,traincode);
        }
    }
};

function SingleLineBox(controller, bell, model) {
    SignalBox.call(this,controller, bell);
    this._model = model;
    this._trainOnTrack = false;
}
Subclass.make(SingleLineBox,SignalBox);

SingleLineBox.prototype.checkDirectionLever = function(need) {
    if (need && this._state === BoxState.AWAITING_THEIR_LINE_CLEAR && this._model.getLever(6).isPulled()) {
        this.changeState(BoxState.PAUSE_TRAIN_START);
        this._scheduler.addGameTimeEvent(SignalBox.pauseTrainDelay,this.sendOutOfOurSection,this);
    } else if (!!!need && this._state === BoxState.AWAITING_THEIR_TRAIN_ON_LINE && !!!this._model.getLever(6).isPulled()) {
        this._trainOnTrack = true;
        this.makeTrainOnLineCall();
        this._scheduler.addRealTimeEvent(1,this.changeState,this,this._preOfferState);
    } else {
        this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,need);
    }
};
SingleLineBox.prototype.exitTrain = function(whatToCallOnceTrainExited,whoToCallOnceTrainExited,whyToProvideOnceTrainExited) {
    if (this._state !== BoxState.EXPECTING_TRAIN) {
        return false;
    }
    if (!!!this._model.is17Free()) {
        this._whatToCallOnceTrainExited = whatToCallOnceTrainExited;
        this._whoToCallOnceTrainExited = whoToCallOnceTrainExited;
        this._whyToProvideOnceTrainExited = whyToProvideOnceTrainExited;
        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
        const bell_length = this._bell.scheduleBellCode(BellCode.Attn,1);
        this._scheduleTimeoutAck();
        return true;
    } else {
        return false;
    }
};
SingleLineBox.prototype.canTrainBeOffered = function() {
    return !!!this._trainOnTrack && this._state === BoxState.NONE;
};
SingleLineBox.prototype.actOnBellPush = function() {
    switch(this._state) {
        case BoxState.NONE:
        case BoxState.PAUSE_AFTER_EXIT_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.RECEIVING);
            }
            break;
        case BoxState.RECEIVING:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                if (this._offeredTrains.length > 0) {
                    const bellLength = this._bell.echoBellCode();
                    this._trainOnTrack = false;
                    this._offeredTrains.shift();
                    this.changeState(BoxState.PAUSE_AFTER_EXIT_ACK);
                    this._scheduler.addRealTimeEvent(bellLength+5,this.changeAfterPause,this);
                } else {
                    this.changeState(BoxState.NONE);
                }
            } else if (BellCode.Cancel.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.NONE);
/* jshint ignore:start */
            } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                // Ignore Bell Code
            } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.release17,this._model);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_SHUNT_ATTN);
            } else {
                // We assume it's an offer - we can only accept if
                //   (a) no offered trains are remaining (i.e. we haven't sent one that still (notionally at least) in section
                //   (b) nothing is on the line
                if (this._offeredTrains.length === 0 &&
                    !!!this._trainOnTrack
                   ) {
                    const bellLength = this._bell.echoBellCode();
                    this._scheduler.addRealTimeEvent(bellLength+1,this._model.release17,this._model);
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_TRAIN_ENTERING);
                }
            }
            break;
        case BoxState.AWAITING_TRAIN_ENTERING:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode()) || 
                BellCode.WronglyDescribed.equals(this._bell.getBellCode()) ||
                BellCode.Cancel.equals(this._bell.getBellCode())
               ) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.lock17,this._model);
                var newState = BoxState.EXPECTING_TRAIN;
                if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) { newState = BoxState.RECEIVING; }
                if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                    this._trainOnTrack = true;
                    this._scheduler.addRealTimeEvent(bellLength+1.005,this.trainIntoSection,this);
                }
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,newState);
            }
            break;
        case BoxState.AWAITING_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OFFER_ACK);
                const bellLength = this._bell.scheduleBellCode(this._trainToOffer,1);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OFFER_ACK:
            if (this._trainToOffer.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_LINE_CLEAR);
                this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,true);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_LINE_CLEAR:
            break;
        case BoxState.PAUSE_TRAIN_START:
            break;
        case BoxState.AWAITING_ENTERING_ACK:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,false);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_TRAIN_ON_LINE:
            break;
        case BoxState.AWAITING_EXIT_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OUT_OF_SECTION_ACK);
                const bellLength = this._bell.scheduleBellCode(BellCode.OutOfSection,1);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OUT_OF_SECTION_ACK:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.PAUSE_AFTER_EXIT_ACK);
                this._scheduler.addRealTimeEvent(5,this.changeAfterPause,this);
                this._trainOnTrack = false;
                this.makeTrainExitedCall();
            }
            break;
        case BoxState.AWAITING_SHUNT_ATTN:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.AWAITING_END_SHUNT);
            }
            break;
        case BoxState.AWAITING_END_SHUNT:
            if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.lock17,this._model);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
            }
            break;
        default:
            break;
    }
};

function DoubleLineBox(controller, bell, ourLine, theirLine) {
    SignalBox.call(this,controller, bell);
    this._ourLine = ourLine;
    this._theirLine = theirLine;
}
Subclass.make(DoubleLineBox,SignalBox);

DoubleLineBox.prototype.checkTheirCommuator = function(need) {
    if (need === CommuatorState.LINE_CLEAR && this._state === BoxState.AWAITING_THEIR_LINE_CLEAR && this._theirLine.getState() === need) {
        this.changeState(BoxState.PAUSE_TRAIN_START);
        this._scheduler.addGameTimeEvent(SignalBox.pauseTrainDelay,this.sendOutOfOurSection,this);
    } else if (need === CommuatorState.TRAIN_ON_LINE && this._state === BoxState.AWAITING_THEIR_TRAIN_ON_LINE && this._theirLine.getState() === need) {
        this.makeTrainOnLineCall();
        this._scheduler.addRealTimeEvent(1,this.changeState,this,this._preOfferState);
    } else if (need === CommuatorState.NORMAL) {
        if (this._theirLine.getState() === CommuatorState.TRAIN_ON_LINE) { // Note: LINE_CLEAR for this case is treated as if it moved through NORMAL first
            if (this._idleStarted && this._scheduler.getClock().getTicks() > this._idleStarted+60) {
                this._messagebox.showMessage(this.getName(),"This is "+this.getName()+" Signaller. I received an 'Train Out of Section' "+
                        "bell code from you, but your commuator is still showing TRAIN ON LINE. Can you move your commuator to NORMAL so "+
                        "that we can get back to standard operation.");
                this._idleStarted = this._scheduler.getClock().getTicks();
            }
            this._scheduler.addGameTimeEvent(3,this.checkTheirCommuator,this,need);
        } else {
            this._idleStarted = null;
            this.changeState(BoxState.PAUSE_AFTER_EXIT_ACK);
            this._scheduler.addRealTimeEvent(5,this.changeAfterPause,this);
        }
    } else {
        this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,need);
    }
};
DoubleLineBox.prototype.exitTrain = function(whatToCallOnceTrainExited,whoToCallOnceTrainExited,whyToProvideOnceTrainExited) {
    if (this._state !== BoxState.NONE) {
        return false;
    }
    if (this._ourLine.getState() === CommuatorState.TRAIN_ON_LINE) {
        this._whatToCallOnceTrainExited = whatToCallOnceTrainExited;
        this._whoToCallOnceTrainExited = whoToCallOnceTrainExited;
        this._whyToProvideOnceTrainExited = whyToProvideOnceTrainExited;
        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
        const bell_length = this._bell.scheduleBellCode(BellCode.Attn,1);
        this._scheduleTimeoutAck();
        return true;
    } else {
        return false;
    }
    
};
DoubleLineBox.prototype.canTrainBeOffered = function() {
    if (this._theirLine.getState() !== CommuatorState.NORMAL) {
        return false;
    }
    return this._state === BoxState.NONE || this._state === BoxState.AWAITING_TRAIN_ENTERING || this._state === BoxState.AWAITING_SHUNT_ATTN;
};
DoubleLineBox.prototype.actOnBellPush = function() {
    switch(this._state) {
        case BoxState.NONE:
        case BoxState.PAUSE_AFTER_EXIT_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.RECEIVING);
            }
            break;
        case BoxState.RECEIVING:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                if (this._offeredTrains.length > 0) {
                    this._bell.echoBellCode();
                    this._offeredTrains.shift();
                    this.startIdleTimer();
                    this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,CommuatorState.NORMAL);
                } else {
                    this.changeState(BoxState.NONE);
                }
            } else if (BellCode.Cancel.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.NONE);
/* jshint ignore:start */
            } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                // Ignore Bell Code
            } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_SHUNT_ATTN);
            } else {
                // We believe it's a train offer - if our commuator is normal we can accept the offer (it'll be TRAIN_ON_LINE if we were expecting
                // attn for out of section
                if (this._ourLine.getState() === CommuatorState.NORMAL) {
                    const bellLength = this._bell.echoBellCode(); 
                    this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.LINE_CLEAR);
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_TRAIN_ENTERING);
                } else {
                    // We're ignoring the offer and hence a new attn will be needed to offer it again
                    this.changeState(BoxState.NONE);
                }
            }
            break;
        case BoxState.AWAITING_TRAIN_ENTERING:
        case BoxState.AWAITING_TRAIN_ENTERING_POSSIBLE_EXIT:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode()) || 
                BellCode.WronglyDescribed.equals(this._bell.getBellCode()) ||
                BellCode.Cancel.equals(this._bell.getBellCode())
               ) {
                const bellLength = this._bell.echoBellCode();
                if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                    this._scheduler.addRealTimeEvent(bellLength+1.005,this.trainIntoSection,this);
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                } else {
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                }
                if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) {
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.RECEIVING);
                } else {
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
                }
            } else if ((this._state === BoxState.AWAITING_TRAIN_ENTERING) && (BellCode.Attn.equals(this._bell.getBellCode()))) {
                this.changeState(BoxState.AWAITING_TRAIN_ENTERING_POSSIBLE_EXIT);
                this._bell.echoBellCode();
            } else if ((this._state === BoxState.AWAITING_TRAIN_ENTERING_POSSIBLE_EXIT) && (BellCode.Attn.equals(this._bell.getBellCode()))) {
                // Repeat of attn, so just echo the attn
                this._bell.echoBellCode();
            } else if ((this._state === BoxState.AWAITING_TRAIN_ENTERING_POSSIBLE_EXIT) && (BellCode.OutOfSection.equals(this._bell.getBellCode()))) {
                if (this._offeredTrains.length > 0) {
                    this._bell.echoBellCode();
                    this._offeredTrains.shift();
                }
                this.changeState(BoxState.AWAITING_TRAIN_ENTERING);
            }
            break;
        case BoxState.AWAITING_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OFFER_ACK);
                this._bell.scheduleBellCode(this._trainToOffer,0);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OFFER_ACK:
            if (this._trainToOffer.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_LINE_CLEAR);
                this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,CommuatorState.LINE_CLEAR);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_LINE_CLEAR:
            break;
        case BoxState.PAUSE_TRAIN_START:
            // We're due to send them an "Entering Section" bell, but they want to something in the meantime
            switch(this._preOfferState) {
                case BoxState.NONE:
                    if (BellCode.Attn.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this._preOfferState = BoxState.RECEIVING;
                    }
                    this.startIdleTimer();
                    break;
                case BoxState.RECEIVING:
                    if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                        if (this._offeredTrains.length > 0) {
                            this._bell.echoBellCode();
                            this._offeredTrains.shift();
                        }
                        this.changePrevState(BoxState.NONE);
                    }
                    if (BellCode.Cancel.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this.changePrevState(BoxState.NONE);
/* jshint ignore:start */
                    } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                        // Ignore Bell Code
                    } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.AWAITING_SHUNT_ATTN);
                    } else {
                        // We believe it's a train offer - if our commuator is normal we can accept the offer (it'll be TRAIN_ON_LINE if we were expecting
                        // attn for out of section
                        if (this._ourLine.getState() === CommuatorState.NORMAL) {
                            const bellLength = this._bell.echoBellCode(); 
                            this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.LINE_CLEAR);
                            this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.AWAITING_TRAIN_ENTERING);
                        } else {
                            // We're ignoring the offer and hence a new attn will be needed to offer it again
                            this.changePrevState(BoxState.NONE);
                        }
                    }
                    break;
                case BoxState.AWAITING_TRAIN_ENTERING:
                    if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                        this._scheduler.addRealTimeEvent(bellLength+1.005,this.expectingTrain,this,true);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.NONE);
                    } else if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.RECEIVING);
                    }
                    break;
                case BoxState.AWAITING_SHUNT_ATTN:
                    if (BellCode.Attn.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this.changePrevState(BoxState.AWAITING_END_SHUNT);
                    }
                    break;
                case BoxState.AWAITING_END_SHUNT:
                    if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.NONE);
                    }
                    break;
                default:
                    break;
            }
            break;
        case BoxState.AWAITING_ENTERING_ACK:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,CommuatorState.TRAIN_ON_LINE);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_EXIT_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OUT_OF_SECTION_ACK);
                const bellLength = this._bell.scheduleBellCode(BellCode.OutOfSection,0);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OUT_OF_SECTION_ACK:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                this._scheduler.addRealTimeEvent(1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                this.changeState(BoxState.PAUSE_AFTER_EXIT_ACK);
                this._scheduler.addRealTimeEvent(5,this.changeAfterPause,this);
                this.makeTrainExitedCall();
            }
            break;
        case BoxState.AWAITING_SHUNT_ATTN:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.AWAITING_END_SHUNT);
            }
            break;
        case BoxState.AWAITING_END_SHUNT:
            if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
            }
            break;
        default:
            break;
    }
};